#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <dos.h>

#define RAND_HALF (RAND_MAX/2)

#define ROWS 25
#define COLS 80

#define BST_OFFR  2
#define BST_OFFC 21

#define VIS_ROWS 10
#define VIS_COLS 10
#define VIS_OFFR  6
#define VIS_OFFC 21

#define ARR_SIZE VIS_ROWS*VIS_COLS

int arr[ARR_SIZE];

struct bstree {
    int num, row, col;
    struct bstree *left, *right;
} *atree;

void bst_insert(struct bstree **tree, int n)
{
    if (*tree == NULL) {
        *tree = malloc(sizeof(struct bstree));
        (*tree)->num = n;
	(*tree)->left = (*tree)->right = NULL;
    }
    else if (n < (*tree)->num) bst_insert(&(*tree)->left, n);
    else if (n > (*tree)->num) bst_insert(&(*tree)->right, n);
    /* else it already exists, so do nothing */
}

void bst_clear(struct bstree **tree)
{
    if (*tree == NULL) return;
    bst_clear(&(*tree)->left);
    bst_clear(&(*tree)->right);
    free(*tree);
    *tree = NULL;
}

int bst_find(struct bstree *tree, int n)
{
    if (tree == NULL) return 0;
    else if (n < tree->num) return bst_find(tree->left, n);
    else if (n > tree->num) return bst_find(tree->right, n);
    else return 1;
}

int bst_find_vis(struct bstree *tree, int n)
{
    if (tree == NULL) return 0;
    gotoxy(tree->col, tree->row);
    highvideo();
    cprintf("%2i", tree->num);
    normvideo();
    gotoxy(1, ROWS);
    delay(500);
    if (n == tree->num) return 1;
    else if (n < tree->num) return bst_find_vis(tree->left, n);
    else if (n > tree->num) return bst_find_vis(tree->right, n);
    /* not reached */
    return 0;
}

void bst_update_pos(struct bstree *tree, int height)
{
    static int col;
    if (height == 0) col = BST_OFFC;
    if (tree == NULL) return;
    bst_update_pos(tree->left, height+1);
    tree->row = height * 2 + BST_OFFR;
    tree->col = col;
    col += 2;
    bst_update_pos(tree->right, height+1);
}

void bst_print(struct bstree *tree)
{
    if (tree == NULL) return;
    gotoxy(tree->col, tree->row);
    cprintf("%2i", tree->num);
    if (tree->left) {
        gotoxy((tree->col+tree->left->col)/2, tree->row+1);
        cprintf("/");
    }
    if (tree->right) {
	gotoxy((tree->col+tree->right->col)/2+1, tree->row+1);
	cprintf("\\");
    }
    bst_print(tree->left);
    bst_print(tree->right);
    gotoxy(1, ROWS);
}

double unirand(void)
{
    return ((double) rand()) / RAND_MAX;
}

void show_search_tree(void)
{
    int i, num;
    bst_clear(&atree);
    for (i = 0; i < 20; i++) bst_insert(&atree, 100 * unirand());
    clrscr();
    bst_update_pos(atree, 0);
    bst_print(atree);
    cprintf("Initializing...\r");
    if (rand() < RAND_HALF) {
	do {
	    num = 100 * unirand();
	}
	while (!bst_find(atree, num));
    }
    else {
	num = 100 * unirand();
	delay(1000);
    }
    cprintf("Searching %2i...\r", num);
    cprintf(bst_find_vis(atree, num) ?
	    "Found!         \r" :
	    "Not Found.     \r");
    getch();
}

void vis_init(void)
{
    int i, j;
    clrscr();
    for (i = 0; i < VIS_ROWS; i++) {
	gotoxy(VIS_OFFC, VIS_OFFR+i);
	for (j = 0; j < VIS_COLS; j++) cprintf("%4i", arr[i*VIS_COLS+j]);
    }
    gotoxy(1, ROWS);
}

void vis_set_current(int pos)
{
    static int old_pos = -1;
    if (pos != -1) {
	gotoxy(VIS_OFFC + (pos % VIS_COLS) * 4, VIS_OFFR + pos / VIS_COLS);
	highvideo();
	cprintf("%4i", arr[pos]);
	normvideo();
    }
    if (pos != -1 && old_pos != -1) {
	gotoxy(VIS_OFFC + (old_pos % VIS_COLS) * 4, VIS_OFFR + old_pos / VIS_COLS);
	cprintf("%4i", arr[old_pos]);
    }
    old_pos = pos;
    gotoxy(1, ROWS);
}

int linear_search(int num)
{
    int i;
    for (i = 0; i < ARR_SIZE; i++) {
       vis_set_current(i);
       delay(200);
       if (arr[i] == num) return 1;
    }
    return 0;
}

int binary_search(int num)
{
    int lo = 0, hi = ARR_SIZE, curr = lo, prev = hi;
    while (curr != prev) {
	prev = curr;
	curr = (lo + hi) / 2;
	vis_set_current(curr);
	delay(500);
	if (arr[curr] == num) return 1;
	else if (arr[curr] > num) hi = curr;
	else if (arr[curr] < num) lo = curr;
    }
    return 0;
}

int group_search(int num)
{
    int i, j;
    for (i = 9; i < ARR_SIZE; i += 5) {
	vis_set_current(i);
	delay(500);
	if (arr[i] == num) return 1;
	else if (arr[i] > num) break;
    }
    if (i >= ARR_SIZE) return 0;
    for (j = i - 1; j > i - 5; j--) {
	vis_set_current(j);
	delay(500);
	if (arr[j] == num) return 1;
	else if (arr[j] < num) break;
    }
    return 0;
}

int intcmp(const void *a, const void *b)
{
    return *((int *) a) - *((int *) b);
}

void show_search_array(int (*do_search)(int))
{
    int num;
    vis_init();
    vis_set_current(-1);
    cprintf("Initializing... \r");
    if (rand() < RAND_HALF) {
	do {
	    num = 1000 * unirand();
	}
	while (!bsearch(&num, arr, ARR_SIZE, sizeof(int), intcmp));
    }
    else {
	num = 1000 * unirand();
	delay(1000);
    }
    cprintf("Searching %3i...\r", num);
    cprintf(do_search(num) ?
	    "Found!          \r" :
	    "Not found.      \r");
    getch();
}

char *menu_options[] = {
    "   Linear search   ",
    "   Binary search   ",
    "   Group search    ",
    "Binary search trees",
    "   Exit program    "
};
int nopt = 5;

int menu(void)
{
    static int curr = 0, oldcurr = 0;
    int i, c, changed = 1;
    clrscr();
    gotoxy(12, 6);
    cprintf("This program demonstrates the most basic searching methods");
    gotoxy(11, 7);
    cprintf("Made by Harris Nikolopoulos 36/91, University Of Macedonia");
    for (i = 0; i < nopt; i++) {
	gotoxy(31, 10+i);
	cprintf("%s", menu_options[i]);
    }
    for (;;) {
        if (changed) {
            gotoxy(28, 10+oldcurr);
            cprintf("  ");
            gotoxy(52, 10+oldcurr);
            cprintf("  ");
            gotoxy(28, 10+curr);
            cprintf("->");
            gotoxy(52, 10+curr);
            cprintf("<-");
            gotoxy(1, ROWS);
            changed = 0;
        }
        c = getch();
        if (c == 13) return curr;
        else if (c == 0) {
            c = getch();
            if (c == 72) {
                oldcurr = curr;
                curr = (curr + nopt - 1) % nopt;
		changed = 1;
            }
            else if (c == 80) {
                oldcurr = curr;
                curr = (curr + 1) % nopt;
                changed = 1;
            }
        }
    }
    /* not reached */
}

int main(void)
{
    int i, num;
    srand(time(NULL));
    rand();
    for (i = 0; i < ARR_SIZE; i++) arr[i] = 1000 * unirand();
    qsort(arr, ARR_SIZE, sizeof(int), intcmp);
    for (;;) switch (menu()) {
      case 0:
	show_search_array(linear_search);
	break;
      case 1:
	show_search_array(binary_search);
	break;
      case 2:
	show_search_array(group_search);
	break;
      case 3:
	show_search_tree();
	break;
      case 4:
        clrscr();
        return 0;
    }
    /* not reached */
}
